@@ -1,13 +1,13 @@
 use strict;
 use warnings;
 package BenchmarkAnything::Storage::Frontend::Lib;
-# git description: v0.003-3-ge09e128
+# git description: v0.004-30-gb054d29
 
 BEGIN {
   $BenchmarkAnything::Storage::Frontend::Lib::AUTHORITY = 'cpan:SCHWIGON';
 }
 # ABSTRACT: Basic functions to access a BenchmarkAnything store
-$BenchmarkAnything::Storage::Frontend::Lib::VERSION = '0.004';
+$BenchmarkAnything::Storage::Frontend::Lib::VERSION = '0.005';
 use Scalar::Util 'reftype';
 
 
@@ -16,6 +16,7 @@ sub new
         my $class = shift;
         my $self  = bless { @_ }, $class;
         $self->_read_config;
+        $self->connect;
         return $self;
 }
 
@@ -184,10 +185,16 @@ sub _read_config
         require File::Slurp;
         require YAML::Any;
 
-        my $configfile     = $self->{cfgfile} || $ENV{BENCHMARKANYTHING_CONFIGFILE} || File::HomeDir->my_home . "/.benchmarkanything.cfg";
-        my $configyaml     = File::Slurp::read_file($configfile);
-        $self->{config}    = YAML::Any::Load($configyaml);
-        $self->{backend} ||= 'tapper';
+        # don't look into user's homedir if we are running tests
+        my $default_cfgfile = $ENV{HARNESS_ACTIVE} ? undef : File::HomeDir->my_home . "/.benchmarkanything.cfg";
+
+        # read file
+        $self->{cfgfile} = $self->{cfgfile} || $ENV{BENCHMARKANYTHING_CONFIGFILE} || $default_cfgfile;
+        $self->{config}  = YAML::Any::Load("".File::Slurp::read_file($self->{cfgfile}));
+
+        # defaults
+        $self->{config}{benchmarkanything}{frontend} ||= 'lib';
+
         return $self;
 }
 
@@ -196,7 +203,8 @@ sub connect
 {
         my ($self) = @_;
 
-        if ($self->{backend} eq 'tapper')
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
         {
                 require DBI;
                 require Tapper::Benchmark;
@@ -214,9 +222,28 @@ sub connect
                 $self->{dbh}              = $dbh;
                 $self->{tapper_benchmark} = Tapper::Benchmark->new({dbh => $dbh, debug => $self->{debug} });
         }
-        else
+        elsif ($frontend eq 'http')
         {
-                die "benchmarkanything: backend ".$self->{backend}." not yet implemented.\nAvailable backends are: 'tapper'\n";
+                my $ua  = $self->_get_user_agent;
+                my $url = $self->_get_base_url."/api/v1/hello";
+                die "benchmarkanything: can't connect to result storage ($url)\n" if (!$ua->get($url)->res->code or $ua->get($url)->res->code != 200);
+        }
+
+        return $self;
+}
+
+
+sub disconnect
+{
+        my ($self) = @_;
+
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
+        {
+                if ($self->{dbh}) {
+                        $self->{dbh}->commit unless $self->{dbh}{AutoCommit};
+                        undef $self->{dbh}; # setting dbh to undef does better cleanup than disconnect();
+                }
         }
         return $self;
 }
@@ -243,7 +270,7 @@ sub _are_you_sure
         }
 
         # ask on stdin
-        print "REALLY DELETE AND RE-CREATE DATABASE [$dsn] (y/N): ";
+        print "REALLY DROP AND RE-CREATE DATABASE TABLES [$dsn] (y/N): ";
         read STDIN, my $answer, 1;
         return 1 if $answer && $answer =~ /^y(es)?$/i;
 
@@ -256,40 +283,33 @@ sub createdb
 {
         my ($self) = @_;
 
-        if ($self->{backend} eq 'tapper')
+        if ($self->_are_you_sure)
         {
-                if ($self->_are_you_sure)
+                no warnings 'once'; # avoid 'Name "DBI::errstr" used only once'
+
+                require DBI;
+                require File::Slurp;
+                require File::ShareDir;
+                use DBIx::MultiStatementDo;
+
+                my $batch            = DBIx::MultiStatementDo->new(dbh => $self->{dbh});
+
+                # get schema SQL according to driver
+                my $dsn      = $self->{config}{benchmarkanything}{backends}{tapper}{benchmark}{dsn};
+                my ($scheme, $driver, $attr_string, $attr_hash, $driver_dsn) = DBI->parse_dsn($dsn)
+                 or die "benchmarkanything: can not parse DBI DSN '$dsn'";
+                my ($dbname) = $driver_dsn =~ m/database=(\w+)/g;
+                my $sql_file = File::ShareDir::dist_file('Tapper-Benchmark', "tapper-benchmark-create-schema.$driver");
+                my $sql      = File::Slurp::read_file($sql_file);
+                $sql         =~ s/^use `testrundb`;/use `$dbname`;/m if $dbname; # replace Tapper::Benchmark's default
+
+                # execute schema SQL
+                my @results = $batch->do($sql);
+                if (not @results)
                 {
-                        no warnings 'once'; # avoid 'Name "DBI::errstr" used only once'
-
-                        require DBI;
-                        require File::Slurp;
-                        require File::ShareDir;
-                        use DBIx::MultiStatementDo;
-
-                        my $batch            = DBIx::MultiStatementDo->new(dbh => $self->{dbh});
-
-                        # get schema SQL according to driver
-                        my $dsn      = $self->{config}{benchmarkanything}{backends}{tapper}{benchmark}{dsn};
-                        my ($scheme, $driver, $attr_string, $attr_hash, $driver_dsn) = DBI->parse_dsn($dsn)
-                         or die "benchmarkanything: can not parse DBI DSN '$dsn'";
-                        my ($dbname) = $driver_dsn =~ m/database=(\w+)/g;
-                        my $sql_file = File::ShareDir::dist_file('Tapper-Benchmark', "tapper-benchmark-create-schema.$driver");
-                        my $sql      = File::Slurp::read_file($sql_file);
-                        $sql         =~ s/^use `testrundb`;/use `$dbname`;/m if $dbname; # replace Tapper::Benchmark's default
-
-                        # execute schema SQL
-                        my @results = $batch->do($sql);
-                        if (not @results)
-                        {
-                                die "benchmarkanything: error while creating BenchmarkAnything DB: ".$batch->dbh->errstr;
-                        }
-
+                        die "benchmarkanything: error while creating BenchmarkAnything DB: ".$batch->dbh->errstr;
                 }
-        }
-        else
-        {
-                die "benchmarkanything: backend ".$self->{backend}." not yet implemented.\nAvailable backends are: 'tapper'\n";
+
         }
 
         return;
@@ -306,55 +326,114 @@ sub add
                 die "benchmarkanything: no input data provided.\n";
         }
 
-        require BenchmarkAnything::Schema;
-        print "Verify schema...\n" if $self->{verbose};
-        if (not my $result = BenchmarkAnything::Schema::valid_json_schema($data))
-        {
+        if (not $self->{skipvalidation}) {
+            require BenchmarkAnything::Schema;
+            print "Verify schema...\n" if $self->{verbose};
+            if (not my $result = BenchmarkAnything::Schema::valid_json_schema($data))
+            {
                 die "benchmarkanything: add: invalid input: ".join("; ", $result->errors)."\n";
+            }
         }
 
-
         # --- add to storage ---
 
-        if ($self->{backend} eq 'tapper')
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
         {
-                # add data
-                print "Add data...\n" if $self->{verbose};
-                foreach my $chunk (@{$data->{BenchmarkAnythingData}}) { # ensure order, because T::Benchmark optimizes multi-chunk entries
-                        my $success = $self->{tapper_benchmark}->add_multi_benchmark([$chunk]);
-                        if (not $success)
+                my $success;
+                if ($self->{queuemode})
+                {
+                        # only queue for later processing
+                        print "Enqueue data [frontend:lib]...\n" if $self->{verbose} or $self->{debug};
+                        $success = $self->{tapper_benchmark}->enqueue_multi_benchmark($data->{BenchmarkAnythingData});
+                }
+                else
+                {
+                        print "Add data [frontend:lib]...\n" if $self->{verbose} or $self->{debug};
+                        # preserve order, otherwise add_multi_benchmark() would reorder to optimize insert
+                        foreach my $chunk (@{$data->{BenchmarkAnythingData}})
                         {
-                                die "benchmarkanything: error while adding data to backend '".$self->{backend}."': ".$@;
+                                print "." if $self->{debug};
+                                $success = $self->{tapper_benchmark}->add_multi_benchmark([$chunk]);
                         }
                 }
-                print "Done.\n" if $self->{verbose};
+                if (not $success)
+                {
+                        die "benchmarkanything: error while adding data: ".$@;
+                }
+                print "Done.\n" if $self->{verbose} or $self->{debug};
+        }
+        elsif ($frontend eq 'http')
+        {
+                my $ua  = $self->_get_user_agent;
+                my $url = $self->_get_base_url."/api/v1/add";
+                print "Add data [frontend:http]...\n" if $self->{verbose} or $self->{debug};
+                my $res = $ua->post($url => json => $data)->res;
+                print "Done.\n" if $self->{verbose} or $self->{debug};
+
+                die "benchmarkanything: ".$res->error->{message}." ($url)\n" if $res->error;
         }
         else
         {
-                die "benchmarkanything: backend ".$self->{backend}." not yet implemented, available backends are: 'tapper'\n";
+                die "benchmarkanything: no frontend '$frontend', available frontends are: 'http', 'lib'.\n";
         }
 
         return $self;
 }
 
+sub _get_user_agent
+{
+        my ($self) = @_;
+
+        require Mojo::UserAgent;
+
+        # query
+        my $ua    = Mojo::UserAgent->new;
+        return $ua;
+}
+
+sub _get_base_url
+{
+        shift->{config}{benchmarkanything}{frontends}{http}{base_url};
+}
+
 
 sub search
 {
-        my ($self, $query) = @_;
+        my ($self, $query, $value_id) = @_;
 
         # --- validate ---
-        if (not $query)
+        if (not $query and not $value_id)
         {
-                die "benchmarkanything: no query data provided.\n";
+                die "benchmarkanything: no query or value_id provided.\n";
         }
 
-        if ($self->{backend} eq 'tapper')
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
         {
+                # single values
+                return $self->{tapper_benchmark}->get_single_benchmark_point($value_id) if $value_id;
                 return $self->{tapper_benchmark}->search_array($query);
         }
+        elsif ($frontend eq 'http')
+        {
+                my $ua  = $self->_get_user_agent;
+                my $url = $self->_get_base_url."/api/v1/search";
+                my $res;
+                if ($value_id) {
+                        $url .= "/$value_id";
+                        $res = $ua->get($url)->res;
+                } else {
+                        $res = $ua->post($url => json => $query)->res;
+                }
+
+                die "benchmarkanything: ".$res->error->{message}." ($url)\n" if $res->error;
+
+                return $res->json;
+        }
         else
         {
-                die "benchmarkanything: backend '.$self->{backend}.' not yet implemented, available backends are: 'tapper'\n";
+                die "benchmarkanything: no frontend '$frontend', available frontends are: 'http', 'lib'.\n";
         }
 }
 
@@ -363,38 +442,72 @@ sub listnames
 {
         my ($self, $pattern) = @_;
 
-        if ($self->{backend} eq "tapper")
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
         {
                 return $self->{tapper_benchmark}->list_benchmark_names(defined($pattern) ? ($pattern) : ());
         }
+        elsif ($frontend eq 'http')
+        {
+                my $ua  = $self->_get_user_agent;
+                my $url = $self->_get_base_url."/api/v1/listnames";
+
+                my $res = $ua->get($url)->res;
+                die "benchmarkanything: ".$res->error->{message}." ($url)\n" if $res->error;
+
+                my $result = $res->json;
+
+                # output
+                return $result;
+        }
         else
         {
-                die "benchmarkanything: backend '.$self->{backend}.' not yet implemented, available backends are: 'tapper'\n";
+                die "benchmarkanything: no frontend '$frontend', available frontends are: 'http', 'lib'.\n";
         }
 }
 
 
+sub gc
+{
+        my ($self) = @_;
+
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
+        {
+                $self->{tapper_benchmark}->gc;
+        }
+}
+
+
+sub process_raw_result_queue
+{
+        my ($self, $count) = @_;
+
+        $count ||= 10;
+
+        my $frontend = $self->{config}{benchmarkanything}{frontend};
+        if ($frontend eq 'lib')
+        {
+                my $dequeued_raw_bench_bundle_id;
+                do {
+                        $dequeued_raw_bench_bundle_id = $self->{tapper_benchmark}->process_queued_multi_benchmark;
+                        $count--;
+                } until ($count < 1 or not defined($dequeued_raw_bench_bundle_id));
+        }
+        else
+        {
+                die "benchmarkanything: only frontend 'lib' allowed here.\n";
+        }
+        return;
+}
+
+
 sub getpoint
 {
         my ($self, $value_id) = @_;
 
-        if ($self->{backend} eq 'tapper')
-        {
-                require DBI;
-                require Tapper::Benchmark;
-
-                # query
-                die "benchmarkanything: please provide a benchmark value_id'\n"
-                 unless $value_id;
-                my $point = $self->{tapper_benchmark}->get_single_benchmark_point($value_id);
-
-                # output
-                return $point;
-        }
-        else
-        {
-                die "benchmarkanything: backend '.$self->{backend}.' not yet implemented, available backends are: 'tapper'\n";
-        }
+        return $self->search(undef, $value_id);
+        die "benchmarkanything: please provide a benchmark value_id'\n" unless $value_id;
 }
 
 sub _multi_point_stats
@@ -527,7 +640,7 @@ sub transform_chartlines
                         my $stdv  = $VALUES{$title}{$x}{stats}{stdv};
                         my $ci95l = $VALUES{$title}{$x}{stats}{ci_95_lower};
                         my $ci95u = $VALUES{$title}{$x}{stats}{ci_95_upper};
-                        say STDERR sprintf("  %-20s . %-7s . avg = %7.2f +- %5.2f (%3d points)", $title, $x, $avg, $stdv, $count) if $verbose;
+                        say STDERR sprintf("  %-20s . %-7s . (ci95l..avg..ci95u) = (%2.2f .. %2.2f .. %2.2f) +- stdv %5.2f (%3d points)", $title, $x, $ci95l, $avg, $ci95u, $stdv, $count) if $verbose;
                 }
         }
 
@@ -583,11 +696,10 @@ Used for critical functions like createdb. Provide a true value or, in
 case of L</createdb>, the DSN of the database that you are about to
 (re-)create.
 
-=item * backend
+=item * skipvalidation
 
-There are potentially multiple different backend stores. Currently
-only backend C<tapper> is supported which means an SQL database
-accessed with C<Tapper::Benchmark|Tapper::Benchmark>.
+Disables schema validation checking, e.g., when you know your data is
+correct and want to save execution time, ususally for C<add()>.
 
 =item * verbose
 
@@ -721,6 +833,12 @@ Connects to the database according to the DB handle from config.
 
 Returns the object to allow chained method calls.
 
+=head2 disconnect
+
+Commits and disconnects the current DB handle from the database.
+
+Returns the object to allow chained method calls.
+
 =head2 _are_you_sure
 
 Internal method.
@@ -759,6 +877,25 @@ Returns an array ref with all metric NAMEs. Optionally allows to
 restrict the search by a SQL LIKE search pattern, allowing C<%> as
 wildcard.
 
+=head2 gc()
+
+Run garbage collector. This cleans up potential garbage that might
+have piled up, in particular qeued raw results that are already
+processed but still in the storage.
+
+Initially the garbage collection is made for the queing functionality
+(see L</process_raw_result_queue> until we are confident it is
+waterproof. However, generally there might be new code arriving in the
+future for which garbage collection might also make sense, so we
+provide this function as general entry point to do The Right Thing -
+whatever that is by that time.
+
+=head2 process_raw_result_queue($count)
+
+Works on the queued entries created by C<add> in I<queuemode=1>. It
+finishes as soon as there are no more unprocessed raw entries, or it
+processed C<$count> entries (default=10).
+
 =head2 getpoint ($value_id)
 
 Returns a single benchmark point with B<all> its key/value pairs.
